<?php

namespace Drupal\subman\Plugin\Block;

use Drupal\Core\Block\BlockBase;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Url;
use Drupal\subman\SubmanSync;
use Drupal\subman\SubmanWebhookException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides a 'SelfserviceBlock' block.
 *
 * @Block(
 *  id = "subman.selfservice_block",
 *  admin_label = @Translation("Selfservice block"),
 * )
 */
class SelfserviceBlock extends BlockBase implements ContainerFactoryPluginInterface {

  const QUERYSTRING_ID_KEY = 'id';

  /**
   * Drupal\subman\SubmanSync definition.
   *
   * @var \Drupal\subman\SubmanSync
   */
  protected $submanSync;

  /**
   * Symfony\Component\HttpFoundation\RequestStack definition.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Constructs a new SelfserviceBlock object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param string $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\subman\SubmanSync $subman_sync
   *   The injected subman.sync service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The injected request_stack service.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    SubmanSync $subman_sync,
    RequestStack $request_stack
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->submanSync = $subman_sync;
    $this->requestStack = $request_stack;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('subman.sync'),
      $container->get('request_stack')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $build = [];
    $items = [];
    $id_valid = FALSE;
    $id = $this->requestStack->getCurrentRequest()->get(self::QUERYSTRING_ID_KEY);
    $original_uri = $this->requestStack->getCurrentRequest()->getUri();

    // Go through all the current user's subscription...
    foreach ($this->submanSync->getSubscriptionsOfSubscriber() as $subscription) {
      $subscription_id = $subscription[SubmanSync::MAP_SUBSCRIBER_NORM][SubmanSync::MAP_SUBSCRIPTION_ID];
      // ... and check whether given id is pointing to it.
      $id_valid = $id_valid || ($id == $subscription_id);
      // ... and add link to subscription to the items list.
      $item_url = Url::fromUri($original_uri, $options = ['query' => [self::QUERYSTRING_ID_KEY => $subscription_id]]);
      $item_text = $this->submanSync->buildSubscriptionTitle($subscription);
      $items[] = [
        '#type' => 'link',
        '#title' => $item_text,
        '#url' => $item_url,
      ];
    }

    // If no ID was given in querystring, and there is only one subscription, use it.
    if (empty($id) && count($items) == 1) {
      $id = $subscription_id;
      $id_valid = TRUE;
    }

    try {
      if (!empty($id)) {
        if ($id_valid) {
          $build['embed'] = $this->submanSync->buildEmbedSelfservice($id);
        }
        else {
          $build = ['#markup' => $this->t('Specified ID is not valid for current user.')];
        }
      }
      elseif (!empty($items)) {
        // $build['embed']['#markup'] = $this->t('No id specified.');
        $build['list'] = [
          '#cache' => ['max-age' => 0],
          'intro' => ['#markup' => $this->t('Choose one of your active subscriptions to view or change:')],
          'item_list' => [
            '#theme' => 'item_list',
            '#wrapper_attributes' => [
              'class' => [
                'subman-selfservice-subscription-list',
              ],
            ],
            '#items' => $items,
          ],
        ];
      }
      else {
        $build = ['#markup' => $this->t('No subscription exists for this user to view or change.')];
      }
    }
    catch (SubmanWebhookException $e) {
      \Drupal::logger('subman')->error('Exception catched to SelfserviceBlock::build(): ' . $e->getMessage());
      $build['#markup'] = $this->t('An error occured.');
    }

    return $build;
  }

}
