<?php

namespace Drupal\subman\Plugin\Block;

use Drupal\subman\SubmanSync;
use Drupal\Core\Block\BlockBase;
use Drupal\subman\SubmanUtilities;
use Drupal\Core\Form\FormStateInterface;
use Drupal\subman\SubmanWebhookException;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a 'SignupBlock' block.
 *
 * @Block(
 *  id = "subman.signup_block",
 *  admin_label = @Translation("Signup block"),
 * )
 */
class SignupBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * Drupal\subman\SubmanSync definition.
   *
   * @var \Drupal\subman\SubmanSync
   */
  protected $submanSync;

  /**
   * Drupal\subman\SubmanUtilities definition.
   *
   * @var \Drupal\subman\SubmanUtilities
   */
  protected $utils;

  /**
   * Symfony\Component\HttpFoundation\RequestStack definition.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Constructs a new SignupBlock object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param string $plugin_definition
   *   The plugin implementation definition.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    SubmanSync $subman_sync,
    SubmanUtilities $utils,
    RequestStack $request_stack
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->submanSync = $subman_sync;
    $this->requestStack = $request_stack;
    $this->utils = $utils;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('subman.sync'),
      $container->get('subman.utilities'),
      $container->get('request_stack')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'subscription_type_id' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $form['subscription_type_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Subscription type ID'),
      '#description' => $this->t('The ID of the subscription (sub-)type to register for. (For Billwerk this is called Plan Variant).'),
      '#default_value' => $this->configuration['subscription_type_id'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $this->configuration['subscription_type_id'] = $form_state->getValue('subscription_type_id');
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $build = [
      '#cache' => ['max-age' => 0],
    ];
    $subscriptionTypeId = $this->configuration['subscription_type_id'];

    try {
      if (!empty($subscriptionTypeId)) {
        $pre_registration_form = $this->utils->config('pre_registration_form', 'Drupal\subman\Form\PreRegistrationForm');
        if (!empty($pre_registration_form)) {
          $build['pre_registration_form'] = \Drupal::formBuilder()->getForm($pre_registration_form, $subscriptionTypeId);
        }
        else {
          $build['embed'] = $this->submanSync->buildEmbedSignup($subscriptionTypeId);
        }
      }
      else {
        $build['embed']['#markup'] = $this->t('No subscription (sub-)type ID specified');
      }
    }
    catch (SubmanWebhookException $e) {
      \Drupal::logger('subman')->error('Exception catched to SignupBlock::build(): ' . $e->getMessage());
      $build['#markup'] = $this->t('An error occured.');
    }

    return $build;
  }

}
