<?php

namespace Drupal\subman;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\subman\Event\SubmanIncomingWebhook;
use Drupal\subman\Event\SubmanSubscriberModified;
use Drupal\subman\Event\SubmanSynchronousResult;
use Drupal\user\Entity\User;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Class SubmanSync.
 *
 * Provides the re-usable parts of handling and syncing data from a
 * subscription management SaaS to Drupal.
 *
 * This is the abstract base class for any specific service that
 * connects to a specific subscription management SaaS.
 * As such, it can and has to be extended by a specific SaaS integration module.
 *
 * As abstract basis, this class provides all parts of external data handling
 * that are independent of any specific SaaS, e.g.:
 * * Work with normalized versions of specific external data
 * * Find and update Drupal user corresponding to normalized external data
 * * Store sync information on to the corresponding Drupal user
 * * Declare abstract base methods (that can be overridden in specific
 *   implementations for specific SaaS).
 *
 * A specific implementation of this base class should be declared as service
 * "subman.sync", and as event subscriber via service tag "event_subscriber".
 * We assume that only one implementation of the abstract class SubmanSync
 * is used at a time. This is an easy way to let other code use a specific
 * service class for a specific SaaS without knowing about it specifically.
 */
abstract class SubmanSync implements SubmanSyncInterface, EventSubscriberInterface {
  use StringTranslationTrait;

  const FIELDNAME_EXTERNAL_ID = 'field_subman_external_id';
  const FIELDNAME_SYNC = 'field_subman_sync';

  const FIELDNAME_ID_PRIMARY = 'field_subman_external_id';
  const FIELDNAME_ID_SECONDARY = 'mail';
  const FIELDNAME_ID_SECONDARY_REMOTE = 'ExternalCustomerId';

  // @todo rename to DATA_COMPONENT_...
  const DATA_KEY_SUBSCRIBER = 'subscriber';
  const DATA_KEY_SUBSCRIPTION = 'subscription';
  const DATA_KEY_SUBSCRIPTION_TYPE = 'subscription_type';
  const DATA_KEY_TIMESTAMP = 'timestamp';

  // @todo Refactor constant name to MAP_NORMALIZED.
  const MAP_SUBSCRIBER_NORM = '_normalized';

  const MAP_SUBSCRIBER_ID = '_id';
  const MAP_SUBSCRIBER_EMAIL = '_email';
  const MAP_SUBSCRIBER_FIRSTNAME = '_firstname';
  const MAP_SUBSCRIBER_LASTNAME = '_lastname';
  const MAP_SUBSCRIBER_FOREIGN_ID = '_foreign_id';
  const MAP_SUBSCRIBER_CREATED = '_created';
  const MAP_SUBSCRIBER_ACTIVE = '_active';
  const MAP_SUBSCRIBER_DELETED = '_deleted';

  const MAP_SUBSCRIPTION_ID = '_id';
  const MAP_SUBSCRIPTION_SUBSCRIBER_ID = '_subscriber_id';
  const MAP_SUBSCRIPTION_TYPE = '_type';
  const MAP_SUBSCRIPTION_VARIANT = '_variant';
  const MAP_SUBSCRIPTION_REFERENCE = '_reference';
  const MAP_SUBSCRIPTION_VALID = '_valid';
  const MAP_SUBSCRIPTION_MESSAGE = '_message';
  const MAP_SUBSCRIPTION_ESCALATION = '_escalation';

  const MAP_SUBSCRIPTION_TYPE_ID = '_id';
  const MAP_SUBSCRIPTION_TYPE_NAME = '_name';
  const MAP_SUBSCRIPTION_TYPE_NAME_TECHNICAL = '_name_technical';
  const MAP_SUBSCRIPTION_TYPE_ROLES = '_roles';
  const MAP_SUBSCRIPTION_TYPE_CLASS = '_class';
  const MAP_SUBSCRIPTION_TYPE_CLASS_MAIN = 'main';
  const MAP_SUBSCRIPTION_TYPE_CLASS_VARIANT = 'variant';

  const NOTIFY_TRIAL_ENDING = 'trial_ending';

  /**
   * Drupal\subman\SubmanUtilities definition.
   *
   * @var \Drupal\subman\SubmanUtilities
   */
  protected $utils;

  /**
   * Constructs a new SubmanSubscriberSync object.
   */
  public function __construct(SubmanUtilities $subman_utilities) {
    $this->utils = $subman_utilities;
  }

  /**
   * Get the title of the subscription management service.
   *
   * @return string
   *   The title of the subscription management service.
   */
  abstract public function getSubscriptionManagementServiceTitle();

  /**
   * {@inheritdoc}
   *
   * @return array
   *   The event names to listen for, and the methods that should be executed.
   */
  public static function getSubscribedEvents() {
    return [
      SubmanIncomingWebhook::EVENT_NAME => 'onIncomingWebhook',
      SubmanSynchronousResult::EVENT_NAME => 'onSynchronousResult',
      SubmanSubscriberModified::EVENT_NAME => 'onSubscriberModified',
    ];
  }

  /**
   * Handle an incoming webhook event.
   *
   * @param \Drupal\subman\Event\SubmanIncomingWebhook $event
   *   The event with the data.
   */
  abstract public function onIncomingWebhook(SubmanIncomingWebhook $event);

  /**
   * Handle a synchronous result event.
   *
   * @param \Drupal\subman\Event\SubmanSynchronousResult $event
   *   The event with the data.
   */
  abstract public function onSynchronousResult(SubmanSynchronousResult $event);

  /**
   * Handle an incoming subscriber modified event.
   *
   * @param \Drupal\subman\Event\SubmanSubscriberModified $event
   *   The event with the subscriber data.
   */
  public function onSubscriberModified(SubmanSubscriberModified $event) {
    // By default, send welcome email with password reset link for newly created
    // subscribers.
    if ($event->operation == SubmanSubscriberModified::OP_CREATED) {
      $this->notifyNewSubscriber($event->subscriber);
    }
  }

  /**
   * Welcome a newly create subscriber.
   *
   * By default this is done by triggering the register_admin_created mail.
   */
  public function notifyNewSubscriber($subscriber) {
    if ($subscriber instanceof User) {
      _user_mail_notify('register_admin_created', $subscriber);
    }
  }

  /**
   * Updates the subscriber by an external id.
   *
   * Gets the customer with given id and triggers subman to update the
   * subscriber with it.
   *
   * @param mixed $external_id
   *   The id of the customer to get updated data on.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function updateSubscriberByExternalId($external_id) {
    // Get customer by id.
    SubmanWebhookException::throwOnEmpty($external_id, 'CustomerId');

    // Get complete data for customer (incl. subscriptions).
    $data = $this->getSubscriberDataComplete($external_id);

    // Do the actual updating.
    $this->updateSubscriberByData($data);
  }

  /**
   * Updates basic subscriber fields with provided data.
   *
   * To extend this updating, override this method in derived/overridden class,
   * with parent::updateSubscriber(...) called last for saving.
   *
   * @param array $data
   *   Array with the subscriber data as it comes from the specific SaaS,
   *   including '_normalized' => ['id' => ..., 'email' => ..., 'active' => bool].
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The content entity to us as subscriber and update with $data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function updateSubscriberByData($data, $subscriber = NULL) {
    $log_data = [];
    $norm = $data[self::DATA_KEY_SUBSCRIBER][self::MAP_SUBSCRIBER_NORM];

    // Lookup subscriber if none is specified.
    $subscriber = $subscriber ?? $this->lookupSubscriber($data[self::DATA_KEY_SUBSCRIBER], TRUE);

    // Abort with exception if no subscriber could be determined by now.
    if (empty($subscriber)) {
      throw new SubmanWebhookException(SubmanWebhookException::subscriberMissOrConflict, $norm[self::MAP_SUBSCRIBER_ID] . ' / ' . ($norm[self::MAP_SUBSCRIBER_EMAIL] ?? ''));
    }

    // Transfer normalized data to fields.
    $this->transferNormalizedDataToFields($subscriber, $norm, $log_data);

    // @todo Move the following three specialized value assignments to config+transferNormalizedDataToFields.
    // Set id / email / active-status, if each data is still present and fields exist.
    if (isset($norm[self::MAP_SUBSCRIBER_ID]) && $subscriber->hasField(self::FIELDNAME_EXTERNAL_ID)) {
      $subscriber->set(self::FIELDNAME_EXTERNAL_ID, $norm[self::MAP_SUBSCRIBER_ID]);
    }
    if (isset($norm[self::MAP_SUBSCRIBER_EMAIL]) && $subscriber->hasField('mail')) {
      $subscriber->set('mail', $norm[self::MAP_SUBSCRIBER_EMAIL]);
    }
    if (isset($norm[self::MAP_SUBSCRIBER_ACTIVE]) && $subscriber->hasField('status')) {
      $subscriber->set('status', $norm[self::MAP_SUBSCRIBER_ACTIVE]);
    }

    // Save sync data into field.
    $this->storeSyncData($subscriber, $data);

    // Update roles.
    $this->assignRolesBySubscriptions($subscriber);

    // Do any additional steps.
    $this->updateSubscriberCustomSteps($data, $subscriber);

    // Save updated entity, if still desirable.
    if (!isset($subscriber->_subman_abort_save)) {
      $isNew = $subscriber->isNew();
      $subscriber->save();

      // Notify others of subscriber change via custom event, e.g. ti send welcome emails etc.
      $event = new SubmanSubscriberModified($subscriber, $isNew ? SubmanSubscriberModified::OP_CREATED : SubmanSubscriberModified::OP_UPDATED);
      $this->utils->dispatchEvent(SubmanSubscriberModified::EVENT_NAME, $event);

      // Log what we did.
      $this->utils->log('Subscriber @entity_id @message_op for external id:@id/email@:email', $log_data, [
        '@entity_id' => $subscriber->id(),
        '@id' => $norm[self::MAP_SUBSCRIBER_ID],
        '@email' => $norm[self::MAP_SUBSCRIBER_EMAIL] ?? $subscriber->get('mail')->value ?? '',
        '@message_op' => $isNew ? 'created' : 'updated',
      ]);
    }
  }

  /**
   * Perform any closing custom steps during updateSubscriberByData()
   * before updated user is saved.
   *
   * This method does nothing by default but can be overridden in a derived and
   * swapped-out service class to be able to add custom steps easily.
   *
   * @param array $data
   *   The sync data as given by caller.
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The user / content entity of the subscriber, incl. current sync data in
   *   field self::FIELDNAME_SYNC.
   */
  public function updateSubscriberCustomSteps(array $data, ContentEntityBase $subscriber) {
  }

  /**
   * Transfers the given normalized data to fields on the given subscriber/user.
   *
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The subscriber/user Drupal entity.
   *
   * @param array $norm
   *   The normalized data as associative array.
   *
   * @param array $log_data
   *   Optional by-reference parameter with log data to ammend as needed.
   */
  public function transferNormalizedDataToFields(ContentEntityBase $subscriber, array $norm, &$log_data = []) {
    // First, set fields defined in extended mapping (disregard "_"defaults).
    foreach ($norm as $key => $value) {
      // @todo clarify and check whether really to check for '_' special fields,
      //   or whether we can also process those by this general routine.
      if (substr($key, 0, 1) != '_') {
        if ($subscriber->hasField($key)) {
          try {
            $subscriber->set($key, $value);
          }
          catch (\Exception $e) {
            $log_data[] = $e->getMessage();
          }
        }
        else {
          $log_data[] = "Field $key not found on entity " . $subscriber->getEntityTypeId() . ' ' . $subscriber->id() . '.';
        }
      }
    }
  }

  /**
   * Gets the subscriber belonging to the given external primary key.
   *
   * In practice, a user is retrieved how carries the given id as its external id.
   *
   * @param $id_primary
   *   string The external primary id to look up against.
   *
   * @return \Drupal\Core\Entity\ContentEntityBase The subscriber entity found.
   */
  public function getSubscriberById($id_primary) {
    $subscriber = $this->utils->retrieveEntityByField('user', [self::FIELDNAME_EXTERNAL_ID => $id_primary]);
    return $subscriber;
  }

  /**
   * Looks up a subscriber entity (here: user) via the provided data.
   *
   * This method assumes the user entity type is used as subscriber.
   * To change the entity type to be used as subscriber, or how to look it up,
   * override this method in derived/overridden service class.
   *
   * @todo Consider and decide whether to make this entity type independent.
   *
   * @param array $data
   *   Array of data, see updateSubscriber().
   * @param bool $create_if_not_found
   *   A flag whether or not to create an empty entity if none found.
   *
   * @return \Drupal\Core\Entity\ContentEntityBase
   *   The subscriber entity found (if any).
   */
  public function lookupSubscriber($data, $create_if_not_found = TRUE) {
    $norm = $this->normalize('subscriber', $data);

    $id_primary = $norm[self::MAP_SUBSCRIBER_ID];
    $id_secondary_drupal = $this->utils->config('mapping.subscriber.secondary_id.drupal', self::FIELDNAME_ID_SECONDARY);
    $id_secondary_remote = $this->utils->config('mapping.subscriber.secondary_id.remote', self::FIELDNAME_ID_SECONDARY_REMOTE);

    // Don't create new user if subscriber is deleted:
    $create_if_not_found = $create_if_not_found & !($norm[self::MAP_SUBSCRIBER_DELETED] ?? FALSE);

    // First approach: try to find a subscriber by id:
    $approach = 1;
    $subscriber = $this->getSubscriberById($id_primary);

    // Second approach: try to find a subscriber by secondary identifier
    // (e.g. email) but only use that user if it has not already a different
    // external id:
    if (!$subscriber) {
      $subscriber = $this->utils->retrieveEntityByField('user', [$id_secondary_drupal => $data[$id_secondary_remote]]);
      // If user found already has a different external id assigned, make sure
      // we return no result. This will lead to an "Subscriber entity could not
      // be found or created" exception being thrown and logged in the caller
      // (e.g. updateSubscriberByData), wherefrom it can be monitored and thus
      // used to trigger an alert. But this will more importantly prevent
      // creation of duplicate users.
      if ($subscriber && !$subscriber->get(self::FIELDNAME_EXTERNAL_ID)->isEmpty() && $subscriber->get(self::FIELDNAME_EXTERNAL_ID)->value != $id_primary) {
        $approach = 2;
        $subscriber = NULL;
      }
    }

    // If no approach succeeded, create empty new user, if desired.
    if (empty($subscriber) && $create_if_not_found && $approach == 1) {
      $subscriber = User::create(['name' => $this->determineUsernameSafe($data)]);
    }

    return $subscriber;
  }

  /**
   * Determine a new user's safe (unique) name.
   *
   * @param array $data
   *   The structured subscriber data array.
   *
   * @return string
   *   The determined username.
   */
  protected function determineUsernameSafe($data) {
    $name = $this->determineUsername($data);

    // If user with same username exists, append part of the external user ID to make username unique.
    if (!empty($this->utils->retrieveEntityByField('user', ['name' => $name]))) {
      $name .= ' ' . substr($data[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_ID], -5);
    }

    return $name;
  }

  /**
   * Determine a new user's name from given data.
   *
   * @param array $data
   *   The structured subscriber data array.
   *
   * @return string
   *   The determined username.
   */
  protected function determineUsername($data) {
    // Assign username as "firstname lastname".
    $name_parts = [
      $data[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_FIRSTNAME],
      $data[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_LASTNAME],
    ];
    $name = implode(' ', $name_parts);

    return $name;
  }

  /**
   * Deletes a user with the given data.
   *
   * @param string $id
   *   The external identifier of the user to be deleted.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function deleteSubscriber($id) {
    // Lookup subscriber.
    $subscriber = $this->lookupSubscriber([self::MAP_SUBSCRIBER_NORM => [self::MAP_SUBSCRIBER_ID => $id]], FALSE);

    // Delete the subscriber.
    if ($subscriber) {
      if ($subscriber->getEntityTypeId() == 'user') {
        user_cancel([], $subscriber->id(), 'user_cancel_delete');
      }
      else {
        $subscriber->delete();
      }
      $this->utils->log('Subscriber @id deleted.', NULL, ['@id' => $id]);
    }
    else {
      throw new SubmanWebhookException(SubmanWebhookException::subscriberMissOrConflict, $id);
    }
  }

  /**
   * Save sync data into field on subscriber.
   *
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The subscriber.
   * @param array $sync
   *   The sync data.
   */
  public function storeSyncData(ContentEntityBase $subscriber, array $sync = []) {
    if ($subscriber->hasField(self::FIELDNAME_SYNC)) {
      $sync[self::DATA_KEY_TIMESTAMP] = $this->utils->getRequestTime();
      $sync = json_encode($sync, JSON_PRETTY_PRINT);
      $subscriber->set(self::FIELDNAME_SYNC, $sync);
    }
  }

  /**
   * Attempts to sync data from the SaaS for and into the given subscriber.
   *
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The user entity to sync with external system.
   * @param bool $force_full_lookup
   *   Whether to force a full subscriber data lookup instead of just using any
   *   external_id in stored sync data (which is the default).
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function syncSubscriber(ContentEntityBase $subscriber, $force_full_lookup = FALSE) {
    $external_id = NULL;

    // First check whether external_id is part of stored sync info.
    if (!$subscriber->get(self::FIELDNAME_SYNC)->isEmpty()) {
      $data = $this->getSyncDataOfUser($subscriber, self::DATA_KEY_SUBSCRIBER);
      $external_id = $this->extractSubscriberIdFromData($data);
    }

    // Alternatively perform a full lookup, if necessary or explicitely called
    // for.
    if (empty($external_id) || $force_full_lookup) {
      $data = $this->lookupSubscriberDataForUser($subscriber);
      $external_id = $this->extractSubscriberIdFromData($data);
    }

    // If external subscriber data found, update local subscriber with data ...
    if ($external_id) {
      $data = $this->getSubscriberDataComplete($external_id);
      $this->updateSubscriberByData($data, $subscriber);
    }
    // ... or with empty sync data/timestamp to avoid future futile sync.
    else {
      $this->storeSyncData($subscriber);
      $subscriber->save();
    }
  }

  /**
   * A wrapper method for the "syncSubscriber()" method.
   *
   * Attempts to sync data from the SaaS for and into the given subscriber,
   * while catching and logging an exception.
   *
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The user entity to sync with external system.
   * @param bool $force_full_lookup
   *   Whether to force a full subscriber data lookup instead of just using any
   *   external_id in stored sync data (which is the default).
   * @param bool $raise_system_message
   *   Wether to raise a system message if the subscriber can't be synced.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function syncSubscriberSafe(ContentEntityBase $subscriber, $force_full_lookup = FALSE, $raise_system_message = FALSE) {
    try {
      $this->syncSubscriber($subscriber, $force_full_lookup);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->utils->log($this->t('Exception :ex_message occured at :ex_position during operation.', [
        ':ex_message' => $e->getMessage(),
        ':ex_position' => $e->getFile() . ':' . $e->getLine(),
      ]), [], ['exception' => $e], 'error');

      if ($raise_system_message) {
        \Drupal::logger('subman')->error('An error occurred, while syncing the subscriber: ' . $subscriber->toLink()->toString());
      }
    }

    return FALSE;
  }

  /**
   * Extracts the subscriber ID from the given sync data structure.
   *
   * @param array $data
   *   Sync data strucure to get the subscriber ID from.
   *
   * @return string|null
   *   The extracted subscriber ID, or NULL if none found.
   */
  public function extractSubscriberIdFromData($data) {
    if (is_array($data) && isset($data[self::MAP_SUBSCRIBER_NORM]) && isset($data[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_ID])) {
      return $data[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_ID];
    }
    else {
      return NULL;
    }
  }

  /**
   * Act on a login of a user.
   *
   * Try a first sync with subscription management service of user hasn't been synced yet.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function handleUserLogin(AccountInterface $account) {
    $user = User::load($account->id());
    if ($user->hasField(self::FIELDNAME_SYNC) && $user->get(self::FIELDNAME_SYNC)->isEmpty()) {
      $this->syncSubscriber($user);
    }
  }

  /**
   * Assign the correct roles to the given user according to its subscriptions.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to update the roles on.
   */
  public function assignRolesBySubscriptions(User $user) {
    // First, remove all subman-managed roles from the user.
    $roles_applicable = $this->utils->config('roles_applicable', ['subscriber']);
    foreach ($roles_applicable as $rid) {
      $user->removeRole($rid);
    }

    // Second, assemble all roles according to the user's valid subscriptions'
    // types.
    $roles = [];
    foreach ($this->getSubscriptionsOfSubscriber($user) as $subscription) {
      $subscription_type_id = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE];
      $subscription_type = $this->getSubscriptionType($subscription_type_id);
      if ($subscription_type) {
        $subscription_type_roles = (array) $subscription_type[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_ROLES];
        $roles = array_merge($roles, $subscription_type_roles);
      }
    }

    // Third, assign the user's roles.
    foreach (array_unique($roles) as $rid) {
      if (in_array($rid, $roles_applicable)) {
        $user->addRole($rid);
      }
    }
  }

  /**
   * Returns an array of subscription data sub-arrays.
   *
   * If desired and by default, only valid subscriptions are returned.
   *
   * @param \Drupal\Core\Entity\ContentEntityBase $subscriber
   *   The content entiy / user that is the subscriber.
   * @param bool $limit_to_valid
   *   Whether or not to filter for only valid subscriptions (true by default).
   *
   * @return array
   *   The array of subscription data sub-arrays.
   */
  public function getSubscriptionsOfSubscriber(ContentEntityBase $subscriber = NULL, $limit_to_valid = TRUE) {
    $user = $this->utils->getCurrentUser($subscriber);
    $result = [];
    $subscriptions = $this->getSyncDataOfUser($user, self::DATA_KEY_SUBSCRIPTION) ?? [];
    foreach ($subscriptions as $subscription) {
      if (!$limit_to_valid || $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_VALID]) {
        $result[] = $subscription;
      }
    }
    return $result;
  }

  /**
   * Gets the desired sync data from the given user object.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to get the sync data from.
   * @param string|null $key
   *   An optional sync data array key to return data for (by default, all sync
   *   data is returned).
   * @param bool $decode_json
   *   Whether to decode the json.
   *
   * @return array
   *   The sync data as associative array.
   */
  public function getSyncDataOfUser(User $user, string $key = NULL, $decode_json = TRUE) {
    $data = $user->get(self::FIELDNAME_SYNC);
    $data = $data ? $data->value : '';
    if (($key || $decode_json) && !empty($data)) {
      $data = json_decode($data, TRUE);
    }
    return empty($key) ? $data : $data[$key] ?? NULL;
  }

  /**
   * Determines the external ID of the given user.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to look at.
   *
   * @return string|null
   *   The external ID or NULL, if none found.
   */
  public function getExternalIdOfUser(User $user) {
    return !$user->get(self::FIELDNAME_EXTERNAL_ID)->isEmpty() ? $user->get(self::FIELDNAME_EXTERNAL_ID)->value : NULL;
  }

  /**
   * Determines whether given or current user is a synced subscriber.
   *
   * @param null $user
   *   Any optionally given user object.
   *
   * @return bool
   *   Whether given or current user is a synced subscriber.
   */
  public function isUserSubscriber($user = NULL) {
    $user = $this->utils->getCurrentUser($user);
    return ($this->getExternalIdOfUser($user) != NULL);
  }

  /**
   * Retrieves the subscriber data from the SaaS that matches the given user.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to match.
   *
   * @return array
   *   The raw external data as array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function retrieveSubscriberDataForUser(User $user) {
    $data = NULL;

    // First approach: look up remote customer by external id field.
    $id = $user->get(self::FIELDNAME_ID_PRIMARY)->value;
    if (!empty(($id))) {
      $data = $this->retrieveSubscriberDataViaExternalKey(trim($id));
    }

    // Second approach: look up remote customer by secondarty criteria
    // (such as email).
    if (empty($data)) {
      $id_secondary = trim($this->getSecondaryId($user));
      if (!empty($id_secondary)) {
        $data = $this->retrieveSubscriberDataViaExternalKey($id_secondary, FALSE);
      }
    }

    return $data;
  }

  /**
   * Get the external subscriber data from given key (primary or secondary id).
   *
   * @param string $key
   *   The key used for external subscriber lookup.
   * @param bool $key_is_primary_id
   *   Indicating whether the given $key is the primary id (or the secondary).
   *
   * @return array|mixed
   *   The subscriber data as nested associative array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  abstract public function retrieveSubscriberDataViaExternalKey($key, $key_is_primary_id = TRUE);

  /**
   * Get and normalize the external subscriber data that matches the given user.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to match.
   *
   * @return array
   *   The raw external data as array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function lookupSubscriberDataForUser(User $user) {
    $data = $this->retrieveSubscriberDataForUser($user);
    $this->normalize(self::DATA_KEY_SUBSCRIBER, $data);
    return $data;
  }

  /**
   * Retrieves the secondary id of the user.
   *
   * Extending classes could override this for e.g. customized checks.
   *
   * @param \Drupal\user\Entity\User $user
   *   The Drupal user to retrieve the secondary id from.
   *
   * @return string|null
   *   The secondary id, or NULL, if no valid secondary id could be determined.
   */
  public function getSecondaryId(User $user) {
    $id_secondary_drupal = $this->utils->config('mapping.subscriber.secondary_id.drupal', self::FIELDNAME_ID_SECONDARY);
    $secondary_id = trim($user->get($id_secondary_drupal)->value);
    return $secondary_id;
  }

  /**
   * Retrieves external subscriber by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscriber to retrieve the subscriber data by.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  abstract public function retrieveSubscriberData($external_id);

  /**
   * Gets and normalizes external subscriber by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscriber to retrieve the subscriber data by.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getSubscriberData($external_id) {
    $data = $this->retrieveSubscriberData($external_id);
    $this->normalize(self::DATA_KEY_SUBSCRIBER, $data);
    return $data;
  }

  /**
   * Get the complete subscriber data.
   *
   * Get and combine all the external data for a subscriber
   * (incl. subscriptions).
   *
   * @param string $external_id
   *   The id of the customer to get updated data on.
   *
   * @return array
   *   The combined subscriber data in the structure:
   *     'subscriber' => [...]
   *     'subscriptions' => [[...], [...]]
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getSubscriberDataComplete($external_id) {
    $data = [
      self::DATA_KEY_SUBSCRIBER => $this->getSubscriberData($external_id),
      self::DATA_KEY_SUBSCRIPTION => $this->getSubscriptionsData($external_id),
    ];

    return $data;
  }

  /**
   * Retrieves external subscription by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscription to retrieve the subscription data by.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  abstract public function retrieveSubscriptionData($external_id);

  /**
   * Gets and normalizes external subscription by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscription to retrieve the subscription data by.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getSubscriptionData($external_id) {
    $data = $this->retrieveSubscriptionData($external_id);
    $this->normalize(self::DATA_KEY_SUBSCRIPTION, $data);
    return $data;
  }

  /**
   * Retrieves external subscriptions by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscriber to retrieve the subscriptions data for.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  abstract public function retrieveSubscriptionsData($external_id);

  /**
   * Retrieves external subscriptions by the given external id.
   *
   * @param string $external_id
   *   The external id for a subscriber to retrieve the subscriptions data for.
   *
   * @return array
   *   The array with the raw external data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getSubscriptionsData($external_id) {
    $data = $this->retrieveSubscriptionsData($external_id);
    $data = $this->normalizeMultiple(self::DATA_KEY_SUBSCRIPTION, $data);
    return $data;
  }

  /**
   * Retrieve data of subscription types from remote service.
   *
   * @return array
   *   The array with the raw external data.
   */
  abstract public function retrieveSubscriptionTypesData();

  /**
   * Get data of a specific subscription type.
   *
   * @param string $type_id
   *   Optional subscription type id to return the data for (by default, all
   *   types are returned as an array).
   *
   * @return array
   *   A subscription type data array for given subscription type,
   *   or array data items of all subscription types data, keyed by subscription
   *   type id.
   */
  public function getSubscriptionType($type_id) {
    if (empty($type_id)) {
      return NULL;
    }

    // Get subscription types (probably from cache).
    $types = $this->getSubscriptionTypes();

    // If not subscription type foud for specified type_id, try again without
    // cache.
    if (!isset($types[$type_id])) {
      $types = $this->getSubscriptionTypes(TRUE);
    }

    return $types[$type_id] ?? NULL;
  }

  /**
   * Get data of all subscription types.
   *
   * @param bool $force_remote_retrieval
   *   Indicates whether or not to force the data retrieval from remote.
   *
   * @return array
   *   Array of data items of all subscription types data, keyed by subscription
   *   type id.
   */
  public function getSubscriptionTypes($force_remote_retrieval = FALSE) {
    // First try if data is cached.
    $types = $this->utils->loadData('subscription_types');

    // If no cache hit, retrieve data and store it in cache as associative
    // array.
    if (empty($types) || $force_remote_retrieval) {
      $types = $this->retrieveSubscriptionTypes();
    }

    return $types;
  }

  /**
   * Freshly retrieves subscription types.
   *
   * @return array
   *   The array of subscription type data, keyed by subscription type id.
   */
  public function retrieveSubscriptionTypes() {
    // Retrieve from remote service.
    $data = $this->retrieveSubscriptionTypesData();
    // Normalize.
    $types = $this->normalizeMultiple('subscription_type', $data);
    // Convert to associative arras keyed by subscription type id.
    $types_assoc = [];
    foreach ($types as $type) {
      $type_id = $type[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_ID];
      $types_assoc[$type_id] = $type;
    }
    $types = $types_assoc;
    // Store associative array.
    $this->utils->saveData('subscription_types', $types);

    return $types;
  }

  /**
   * Normalizes the given data for the given type as defined by config.
   *
   * You can define config based normalization for TYPE via:
   *   $settings['subman.settings']['mapping.TYPE'] = [
   *     'field_city.value' => 'Address.City',
   *   ];
   *
   * @param string $type
   *   The type to normalize (e.g. subscriber or subscription).
   *
   * @param array $data
   *   Customer data array as received by the subscription management service.
   *
   * @return array
   *   Array with normalized only data structured as following:
   *     'id'      => ... // String: The id of the subscriber
   *     'email'   => ... // String: the email address of the subscriber
   *     'created' => ... // Int: unix timestamp of the creation date of the
   *     subscriber
   *     'active'  => ... // Boolean: Whether the subscriber is active
   *     'deleted' => ... // Boolean: Whether the subscriber is deleted
   */
  public function getNormalizedData($type, $data) {
    $normalized = [];

    // Get mapping.
    $mapping = $this->utils->config("mapping.$type") ?? [];

    // For each mapping definition...
    foreach ($mapping as $norm_key => $orig_key) {
      // ... traverse original data structure to retrieve original value ...
      $orig_value = $data;
      foreach (explode('.', $orig_key) as $orig_key_part) {
        if (isset($orig_value[$orig_key_part])) {
          $orig_value = $orig_value[$orig_key_part];
        }
        else {
          $orig_value = NULL;
          break;
        }
      }
      // ... and assign original value to the normalized key.
      if ($orig_value != NULL) {
        $normalized[$norm_key] = $orig_value;
      }
    }

    return $normalized;
  }

  /**
   * Adds normalized form for and to given data.
   *
   * For more details see ::getNormalizeData().
   *
   * @param string $type
   *   The type to normalize (e.g. subscriber or subscription).
   * @param array $data
   *   Customer data array as received by the subscription management service.
   */
  public function normalize($type, &$data) {
    if (!isset($data[self::MAP_SUBSCRIBER_NORM])) {
      $data[self::MAP_SUBSCRIBER_NORM] = $this->getNormalizedData($type, $data);
    }
    return $data[self::MAP_SUBSCRIBER_NORM];
  }

  /**
   * Normalize an array of multiple data items of given type.
   *
   * @param string $type
   *   The data type key to normalize for.
   * @param $data
   *   The array of data items (each being nested array itself).
   *
   * @return array
   *   The array of data items now augmented with normalized data.
   */
  public function normalizeMultiple($type, $data) {
    $normalized = [];
    foreach ((array) $data as $data_item) {
      $this->normalize($type, $data_item);
      $normalized[] = $data_item;
    }
    return $normalized;
  }

  /**
   * Builds a title for output of the given subscription data.
   *
   * @param array $subscription
   *   The subscription data as structured array.
   *
   * @return string
   *   The assembled title for the subscription.
   */
  public function buildSubscriptionTitle(array $subscription) {
    $subscription_id = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_ID];
    $subscription_reference = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_REFERENCE];
    $subscription_variant_id = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_VARIANT];
    $subscription_variant = $this->getSubscriptionType($subscription_variant_id);
    $subscription_type_name = $subscription_variant[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_NAME];
    return ($subscription_reference ?? $subscription_id) . ' – ' . $subscription_type_name;
  }

  /**
   * Builds and returns a render array element with an iframe to embed the $url.
   *
   * @param string $url
   *   The URL to show on the iframe embed.
   *
   * @return array
   *   Render array for displaying the iframe embed.
   */
  public function buildEmbed($url) {
    $build = [
      '#type' => 'inline_template',
      '#template' => '<iframe class="subman-embed" src="{{ url }}"></iframe>',
      '#cache' => ['max-age' => 0],
      '#context' => [
        'url' => $url,
      ],
      '#attached' => [
        'library' => 'subman/iframe_embed',
      ],
    ];

    return $build;
  }

  /**
   * Builds and returns render array element to embed a signup form via iframe.
   *
   * @param string $embed_parameter
   *   The URL parameters to extend to the URL.
   *
   * @return array
   *   The render array element to display the iframe embed.
   */
  public function buildEmbedSignup($embed_parameter) {
    return [];
  }

  /**
   * Build embedded self service.
   *
   * Builds and returns render array element to embed a self-service form via
   * iframe.
   *
   * @param string $embed_parameter
   *   The URL parameters to extend to the URL.
   *
   * @return array
   *   The render array element to display the iframe embed.
   */
  public function buildEmbedSelfservice($embed_parameter) {
    return [];
  }

  /**
   * Send the notification of given key to given user.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to send the notification to.
   * @param string $notification_key
   *   The notification to send. Can be configured via:
   *     $settings['subman.settings']['notifications']['trial_ending'] = [
   *       'subject' => 'A notification from [site:name]',
   *       'text' => "Your subscription on [site:name] is there for you.",
   *     ];.
   */
  public function notifyUser($user, $notification_key) {
    $notifications = $this->utils->config('notifications', [
      self::NOTIFY_TRIAL_ENDING => [
        'subject' => 'Your trial phase is ending soon',
        'text' => "Hello [user:display-name],\n\nthe trial phase of your subscription for [site:name] is ending soon.\n\nYou can always view and edit your subscription when you are logged in via [site:login-url].\n\nYour [site:name] Team",
      ],
    ]);

    $notification = $notifications[$notification_key] ?? NULL;

    if (!empty($notification) && isset($notification['text'])) {
      // Translate template, subject and text.
      $subject = $this->t($notification['subject'], [], ['langcode' => $user->getPreferredLangcode()]);
      $text    = $this->t($notification['text'], [], ['langcode' => $user->getPreferredLangcode()]);

      // Replace tokens using $user for [user].
      $token_service = \Drupal::token();
      $data          = ['user' => $user];
      $subject       = $token_service->replace($subject, $data, ['clear' => TRUE]);
      $text          = $token_service->replace($text, $data, ['clear' => TRUE]);

      // Send notification.
      $result = $this->notifyUserSend($user, $subject, $text, $notification_key);

      // Log any fail.
      if (!$result) {
        $this->utils->log('Failed sending notification @notification_key to @user', [], [
          '@notification_key' => $notification_key,
          '@user' => $user->id(),
        ], 'error');
      }
    }
  }

  /**
   * Actually send a notification.
   *
   * By default, mail is used. To use a different notification channel,
   * override this method in a derived and swapped out service class.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to notify.
   * @param string $subject
   *   The subject text of the notification.
   * @param string $text
   *   The body text of the notification.
   * @param string $notification_key
   *   The key specifying of the notification.
   *
   * @return bool
   *   Whether the operation was successful (true) or not (false).
   */
  public function notifyUserSend($user, $subject, $text, $notification_key) {
    $mailManager = \Drupal::service('plugin.manager.mail');
    $module = 'subman';
    $key = 'notification';
    $to = $user->getEmail();
    $langcode = $user->getPreferredLangcode();
    $params['notification_subject'] = $subject;
    $params['notification_text'] = $text;
    $params['notification_key'] = $notification_key;
    $send = TRUE;
    $result = $mailManager->mail($module, $key, $to, $langcode, $params, NULL, $send);
    return (($result['result'] ?? FALSE) === TRUE);
  }

  /**
   * Prepare any action to perform on the given user.
   *
   * @param array $event
   *   The event of the external webhook event that triggered this.
   * @param \Drupal\user\Entity\User $user
   *   The user affected.
   */
  public function performAdditionalWebhookAction($event, $user) {
    // Here might be some actions in the future.
    // Custom derived/extended service classes might override this method.
  }

}
